/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
accdb/accdb.c - ACCDB frontend
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <ctype.h>
#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#ifdef _WIN32
#    define dlopen(filename, flags) LoadLibrary(filename)
#    define dlsym(handle, symbol) GetProcAddress((handle), (symbol))
#    define dlclose(handle) FreeLibrary(handle)
#endif
#define MAXFNLEN 1024 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

struct opt {
    char *default_backend;
};

static int genpw_random(char *, size_t, unsigned long, void *);
static int genpw_phonemic(char *, size_t, unsigned long, void *);
static void read_config(struct opt *, struct rconfig_opt *);

//-----------------------------------------------------------------------------
struct accdb_module *accdb_load(char *module) {
    const char *ext[] = {"", ".so", ".dll", NULL}, **extp = ext;
    struct accdb_module *mp = NULL;
    struct opt config;
    int our_errno = 0;
    struct rconfig_opt table[] = {
        {"DEFAULT_BACKEND", 's', &config.default_backend, NULL},
        {NULL},
    };

    memset(&config, 0, sizeof(struct opt));
    read_config(&config, table);

    if(strcmp(module, "*") == 0 && (module = config.default_backend) == NULL) {
        our_errno = errno;
        goto __accdb_open__err_noopen;
    }

    if((mp = malloc(sizeof(struct accdb_module))) == NULL) {
        our_errno = errno;
        goto __accdb_open__err_noopen;
    }

    // Try plain filename first (as does accdbinfo)
    if((mp->handle = dlopen(module, RTLD_NOW)) == NULL) {
        // If not, prepend accdb_, append an extension and try again
        while(*extp != NULL) {
            char fn[MAXFNLEN];
            snprintf(fn, MAXFNLEN, "accdb_%s%s", module, *extp);

            /* -Wl,-rpath,. makes the dynamic linker search paths at runtime,
            no need to repeat that here. */
            if((mp->handle = dlopen(fn, RTLD_NOW)) != NULL) { break; }
            ++extp;
        }
    }

    if(mp->handle == NULL) { goto __accdb_open__err_noopen; }

    mp->name      = dlsym(mp->handle, "_module_name");
    mp->desc      = dlsym(mp->handle, "_module_desc");
    mp->info      = dlsym(mp->handle, "_module_info");
    mp->open      = dlsym(mp->handle, "q_open");
    mp->usertrav  = dlsym(mp->handle, "q_usertrav");
    mp->userinfo  = dlsym(mp->handle, "q_userinfo");
    mp->useradd   = dlsym(mp->handle, "q_useradd");
    mp->usermod   = dlsym(mp->handle, "q_usermod");
    mp->userdel   = dlsym(mp->handle, "q_userdel");
    mp->grouptrav = dlsym(mp->handle, "q_grouptrav");
    mp->groupinfo = dlsym(mp->handle, "q_groupinfo");
    mp->groupadd  = dlsym(mp->handle, "q_groupadd");
    mp->groupmod  = dlsym(mp->handle, "q_groupmod");
    mp->groupdel  = dlsym(mp->handle, "q_groupdel");
    mp->close     = dlsym(mp->handle, "q_close");
    mp->sysctl    = dlsym(mp->handle, "p_sysctl");

    if(mp->open == NULL) {
        our_errno = EFAULT;
        goto __accdb_open__err_noopen;
    }

    accdb_rconfig_free(table);
    return mp;

__accdb_open__err_noopen:
    accdb_rconfig_free(table);
    free(mp);
    errno = our_errno;
    return NULL;
}

void accdb_unload(struct accdb_module *mp) {
    dlclose(mp->handle);
    free(mp);
    return;
}

int accdb_genpw(char *plain, char **crypted, size_t len, unsigned long meth,
 unsigned long flags, void *randp) {
//  char *pool = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz"; // salt pool

    if(flags & PW_PHONEMIC) {
        genpw_phonemic(plain, len, flags & ~PW_PHONEMIC, randp);
    } else {
        genpw_random(plain, len, flags, randp);
    }

    if(crypted == NULL) { return 1; }

/*  char *ptp, *salt, *saltp;
  size_t i = s->len;
  if((s->pt = malloc(i + 1)) == NULL) { return NULL; }
  if((salt = malloc(32)) == NULL) {
    free(s->pt);
    return NULL;
  }

  ptp = s->pt;

  *ptp = '\0';
  switch(s->method) {
    case PW_DES:
      saltp = salt;
      i = 2;
      while(i--) {
       *saltp++ = pool[(int)((double)HX_rand(randp) * 64 / RAND_MAX)]; }
      *saltp = '\0';
      break;
    case PW_MD5:
      strcpy(salt, "$1$");
      saltp = salt + 3;
      i = 8;
      while(i--) { *saltp++ = pool[(int)((double)HX_rand(randp) * 64 / RAND_MAX)]; }
      *saltp = '\0';
      break;
    case PW_BLOWFISH:
      strcpy(salt, "$2a$05$");
      saltp = salt + 7;
      i = 22;
      while(i--) { *saltp++ = pool[(int)((double)HX_rand(randp) * 64 / RAND_MAX)]; }
      *saltp = '\0';
      break;
  }

  s->cryp = HX_strdup(crypt(s->pt, salt));
  free(salt);
  return s->cryp;*/
  return 0;
}

//-----------------------------------------------------------------------------
static int genpw_random(char *plain, size_t size, unsigned long fl, void *rd) {
    size_t cnt = size;
    char *pcase = NULL, *ptrav = plain;

    while(cnt--) {
        // 20% probability for a digit
        *ptrav++ = !(int)((double)HX_rand(rd) * 5 / RAND_MAX) ?
         '0' + (int)((double)HX_rand(rd) * 10 / RAND_MAX) :
         'a' + (int)((double)HX_rand(rd) * 26 / RAND_MAX);
    }

    if(fl & PW_ONE_CASE) {
        pcase = &plain[(int)((double)HX_rand(rd) * size / RAND_MAX)];
        *pcase = toupper(*pcase);
    }

    if(fl & PW_ONE_DIGIT) {
        char *pdigit;
        while((pdigit = &plain[(int)((double)HX_rand(rd) *
         size / RAND_MAX)]) == pcase); // do not digitize the case letter
        *pdigit = '0' + (int)((double)HX_rand(rd) * 10 / RAND_MAX);
    }

    return 1;
}

/*-----------------------------------------------------------------------------
Password generation function utilizing phonemic rules to generate secure,
but pronouncible and easy-memorizable passwords.

From ftp://ftp.gwdg.de/pub/suse/i386/8.2/suse/src/pwgen-2.02-68.src.rpm

Copyright (C) 2001, 2002 by Theodore Ts'o <tytso@alum.mit.edu>
This file [part] may be distributed under the terms of the GNU Public License.

The code was taken from pwgen-2.02/pw_phonemes.c and has been incorporated into
here and also altered. The original author has nothing to do with this here, so
do not bug him if you find a bug. */

static int genpw_phonemic(char *buf, size_t size, unsigned long ft_flags,
 void *randp) {
    const struct pw_element {
        char *str;
        enum {
            FL_CONSONANT = 1 << 0,
            FL_VOWEL     = 1 << 1,
            FL_DIPTHONG  = 1 << 2,
            FL_NOT_FIRST = 1 << 3,
        } flags;
    } phonemic_element[] = {
        {"a",  FL_VOWEL},
        {"ae", FL_DIPTHONG | FL_VOWEL},
        {"ah", FL_DIPTHONG | FL_VOWEL},
        {"ai", FL_DIPTHONG | FL_VOWEL},
        {"b",  FL_CONSONANT},
        {"c",  FL_CONSONANT},
        {"ch", FL_CONSONANT | FL_DIPTHONG},
        {"d",  FL_CONSONANT},
        {"e",  FL_VOWEL},
        {"ee", FL_DIPTHONG | FL_VOWEL},
        {"ei", FL_DIPTHONG | FL_VOWEL},
        {"f",  FL_CONSONANT},
        {"g",  FL_CONSONANT},
        {"gh", FL_CONSONANT | FL_DIPTHONG | FL_NOT_FIRST},
        {"h",  FL_CONSONANT},
        {"i",  FL_VOWEL},
        {"ie", FL_DIPTHONG | FL_VOWEL},
        {"j",  FL_CONSONANT},
        {"k",  FL_CONSONANT},
        {"l",  FL_CONSONANT},
        {"m",  FL_CONSONANT},
        {"n",  FL_CONSONANT},
        {"ng", FL_CONSONANT | FL_DIPTHONG | FL_NOT_FIRST},
        {"o",  FL_VOWEL},
        {"oh", FL_DIPTHONG | FL_VOWEL},
        {"oo", FL_DIPTHONG | FL_VOWEL},
        {"p",  FL_CONSONANT},
        {"ph", FL_CONSONANT | FL_DIPTHONG},
        {"qu", FL_CONSONANT | FL_DIPTHONG},
        {"r",  FL_CONSONANT},
        {"s",  FL_CONSONANT},
        {"sh", FL_CONSONANT | FL_DIPTHONG},
        {"t",  FL_CONSONANT},
        {"th", FL_CONSONANT | FL_DIPTHONG},
        {"u",  FL_VOWEL},
        {"v",  FL_CONSONANT},
        {"w",  FL_CONSONANT},
        {"x",  FL_CONSONANT},
        {"y",  FL_CONSONANT},
        {"z",  FL_CONSONANT},
    };
#define NUM_ELEMENTS (sizeof(phonemic_element) / sizeof(struct pw_element))
#define randnum(__number) (int)((double)HX_rand(randp) * (__number) / RAND_MAX)

    unsigned long flags, prev, should_be;
    size_t c, len;
    int first, i;
    char *str;

try_again:
    c = 0;
    prev = 0x0;
    should_be = 0;
    first = 1;

    should_be = randnum(2) ? FL_VOWEL : FL_CONSONANT;
  
    while(c < size) {
        i = randnum(NUM_ELEMENTS);
        str = phonemic_element[i].str;
        len = strlen(str);
        flags = phonemic_element[i].flags;

        // Filter on the basic type of the next element
        if((flags & should_be) == 0) { continue; }

        // Handle the FL_NOT_FIRST flag
        if(first && (flags & FL_NOT_FIRST)) { continue; }

        // Do not allow FL_VOWEL followed a FL_VOWEL/FL_DIPTHONG pair
        if((prev & FL_VOWEL) && (flags & FL_VOWEL) && (flags & FL_DIPTHONG)) {
            continue;
        }

        // Do not allow us to overflow the buffer
        if(len > size - c) { continue; }

        // OK, we found an element which matches our criteria, let 's do it!
        strcpy(buf + c, str);

        /* Handle PW_ONE_CASE */
        if(ft_flags & PW_ONE_CASE) {
            if((first || flags & FL_CONSONANT) && randnum(10) < 3) {
                buf[c] = toupper(buf[c]);
                ft_flags &= ~PW_ONE_CASE;
            }
        }

        c += len;

        // Time to stop?
        if(c >= size) { break; }

        // Handle PW_ONE_DIGIT
        if(ft_flags & PW_ONE_DIGIT) {
            if(!first && randnum(10) < 3) {
                buf[c++] = randnum(10) + '0';
                buf[c] = '\0';
                ft_flags &= ~PW_ONE_DIGIT;

                first = 1;
                prev = 0x0;
                should_be = randnum(2) ? FL_VOWEL : FL_CONSONANT;
                continue;
            }
        }

        // OK, figure out what the next element should be
        if(should_be == FL_CONSONANT) {
            should_be = FL_VOWEL;
        } else {
            // should_be == FL_VOWEL
            if(prev & FL_VOWEL || flags & FL_DIPTHONG || randnum(10) > 3) {
                should_be = FL_CONSONANT;
            } else {
                should_be = FL_VOWEL;
            }
        }

        prev = flags;
        first = 0;
    }

    if(ft_flags & (PW_ONE_CASE | PW_ONE_DIGIT)) { goto try_again; }
    return 1;
#undef NUM_ELEMENTS
#undef randnum
}

//-----------------------------------------------------------------------------
static void read_config(struct opt *config, struct rconfig_opt *table) {
    const char *paths[] = {"etc", "/etc", NULL};
    accdb_rconfig_pv(paths, "libaccdb/accdb", table, RCONFIG_ONE);
    return;
}

//==[ End of file ]============================================================
