/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
accdb/rconfig.c - Read simple key-value configuration files
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#define MAXFNLEN 1024 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

/*
Multiple possibilities for rconfig_pv()'s flags:

RCONFIG_ONE
  Parsing will stop after one file has been successfully parsed. This allows
  for a "hardcoded+personal" configuration style. If you had

    const char **pv = {".", "/usr/local/etc", "/etc", NULL};
    rconfig_pv(pv, "app.conf", table, flags);

  only the first (from left to right) successfully parsed will be read.

  To build the current options from default files "hardcoded+default+personal",
  this would be the answer:

    const char **pv = {"/etc", "/usr/local/etc", home_dir, ".", NULL};
    rconfig_pv(pv, "app.conf", table, flags);

  which reads all files, from left to right.
*/

//-----------------------------------------------------------------------------
int accdb_rconfig(char *file, struct rconfig_opt *tab) {
    FILE *fp;
    char buf[MAXLNLEN];

    if((fp = fopen(file, "r")) == NULL) { return -errno; }
  
    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        struct rconfig_opt *travp = tab;
        char *qkey = buf, *qval;

        if(*buf == '#') { continue; }
        HX_chomp(buf);
        if(*buf == '\0') { continue; }
        if((qval = strchr(buf, '=')) == NULL) { continue; }
        *qval++ = '\0';

        while(travp->key != NULL) {
            if(strcmp(qkey, travp->key) == 0) {
                switch(travp->type) {
                    case 'B':
                        *(byte *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case 'H':
                        *(unsigned short *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case 'I':
                        *(unsigned int *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case 'L':
                        *(unsigned long *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case 'b':
                        if(strcmp(qval, "yes") == 0 || strcmp(qval, "1") == 0) {
                          *(int *)travp->ptr = 1;
                        } else {
                          *(int *)travp->ptr = 0;
                        }
                        break;
                    case 'c':
                        *(char *)travp->ptr = *qval;
                        break;
                    case 'd':
                        *(double *)travp->ptr = strtod(qval, NULL);
                        break;
                    case 'f':
                        *(float *)travp->ptr = strtod(qval, NULL);
                        break;
                    case 'h':
                        *(short *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case 'i':
                        *(int *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case 'l':
                        *(long *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case 's':
                        *(char **)travp->ptr = HX_strdup(qval);
                        break;
                }

                if(travp->callback != NULL) {
                    travp->callback(travp->key, travp->type, travp->ptr);
                }
            }
            ++travp;
        }
    }

    fclose(fp);
    return 1;
}

void accdb_rconfig_pv(const char **path, char *file, struct rconfig_opt *tab,
 unsigned long flags) {
    char buf[MAXFNLEN];

    while(*path != NULL) {
        int rv;
        snprintf(buf, MAXLNLEN, "%s/%s", *path, file);
        rv = accdb_rconfig(buf, tab);
        if(flags & RCONFIG_ONE && rv) { break; }
        ++path;
    }

    return;
}

void accdb_rconfig_free(struct rconfig_opt *tab) {
    while(tab->key != NULL) {
        char **ptr = tab->ptr;
        if(tab->type == 's' && *ptr != NULL) { free(*ptr); }
        ++tab;
    }
    return;
}

//==[ End of file ]============================================================
