/*=============================================================================
vitalnix/dxt - User Management UI
ds_xml.c - XML data source module
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
/* This module is documented in some places. More information can be found in
doc/uds_mod.html and doc/uds_xml.html, including the tags recognized. */
#include <libxml/parser.h>
#include <errno.h>
#include <iconv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dxt-src/ds.h"

typedef unsigned char byte;

MODULE_NAME("XML data source module");
MODULE_DESC("no description");
MODULE_TYPE("DS");

struct state {
  xmlDoc *doc;
  xmlNode *ptr;
  iconv_t cd;
};

int q_open(char *, void **);
int q_read(void *, struct ds_entry *);
void q_close(void *);

static char *convert(iconv_t, char *);

/*-----------------------------------------------------------------------------
"The XML data source module as an example" (uds_mod.html):

Data Source modules do just read contents and return them back. q_open() of the
XML DS module calls some libxml initialization routines and xmlParseFile()
which in turn lexes and parses the file structure down. The libxml data
pointers are stored in the state struct, which we can define and which is
passed to every function.

q_read() is then called with the state and a destination struct ds_entry. The
XML file gets read (in fact, it already has been, by xmlParseFile()) -- well
the elements get captured with xmlGetProp() and stuffed into the ds_entry.

After all is finished, q_close() is invoked to free the XML parser data, and
clean things up. As easy as that. If an error happened in q_open(), q_close()
is also called, intercepts what has been initialized so far and only
deallocates valid stuff. */

//-----------------------------------------------------------------------------
int q_open(char *fn, void **state) {
  struct state *sp;
  if((sp = *state = malloc(sizeof(struct state))) == NULL) { return -errno; }

  /* Need iconv() since libxml stores everything in UTF-8 internally, and we
  need it converted back, since Daxtraq wants ISO-8859-1 (Latin-1). */
  sp->cd = iconv_open("iso-8859-1", "utf-8");
  xmlKeepBlanksDefault(0);

  /* xmlParseFile() already scans the file completely and stores the structure
  tree plus components in memory. */
  if((sp->doc = xmlParseFile(fn)) == NULL) {
    q_close(*state);
    return -EIO;
  }

  if((sp->ptr = xmlDocGetRootElement(sp->doc)) == NULL ||
   strcmp(sp->ptr->name, "datasource") != 0) {
    q_close(*state);
    return 0;
  }

  sp->ptr = sp->ptr->children;
  return 1;
}

int q_read(void *state, struct ds_entry *x) {
  /* The reading process is very simple, since libxml takes all the pain to
  parse the file down, and its API directly provides us with the stuff we
  want. */
  struct state *sp = state;
  if(sp->ptr == NULL) { return 0; }
  if(strcmp(sp->ptr->name, "user") == 0) {
    x->vname = convert(sp->cd, xmlGetProp(sp->ptr, "vname"));

    if(x->vname == NULL) { // bogus entry
      sp->ptr = sp->ptr->next;
      return q_read(sp, x);
    }

    x->nname = convert(sp->cd, xmlGetProp(sp->ptr, "nname"));
    x->sgroup = convert(sp->cd, xmlGetProp(sp->ptr, "sgroup"));
    x->xuid_in = convert(sp->cd, xmlGetProp(sp->ptr, "xuid"));
  }

  sp->ptr = sp->ptr->next;
  return 1;
}

void q_close(void *state) {
  struct state *sp = state;
  if(sp->doc != NULL) { xmlFreeDoc(sp->doc); }
  xmlCleanupParser();
  iconv_close(sp->cd);
  free(sp);
  return;
}

//-----------------------------------------------------------------------------
static char *convert(iconv_t cd, char *in) {
  char *inp = in, *out, *outp;
  size_t inl, outl;
  if(in == NULL) { return NULL; }
  if((outp = out = malloc(inl = outl = strlen(in) + 1)) == NULL) {
    return NULL;
  }
  iconv(cd, &inp, &inl, &outp, &outl);
  xmlFree(in);
  return out;
}

//==[ End of file ]============================================================
